from datetime import datetime
from pymongo.monitoring import CommandListener
from opentelemetry.trace import StatusCode

from blocks_genesis.auth.blocks_context import BlocksContextManager
from blocks_genesis.lmt.activity import Activity


class MongoEventSubscriber(CommandListener):
    def __init__(self):
        self._activities = {}

    def started(self, event):
        activity = Activity.start(f"MongoDb::{event.command_name}")
        activity.set_properties({
            "db.system": "mongodb",
            "db.operation": event.command_name,
            "db.request_id": str(event.request_id),
            "db.timestamp": datetime.now().isoformat(),
            "baggage.TenantId": BlocksContextManager.get_context().tenant_id if BlocksContextManager.get_context() else "miscellaneous"
        })

        self._activities[event.request_id] = activity

    def succeeded(self, event):
        activity = self._activities.pop(event.request_id, None)
        if activity:
            activity.set_property("db.duration_ms", event.duration_micros / 1000)
            activity.set_status(StatusCode.OK)
            activity.stop()

    def failed(self, event):
        activity = self._activities.pop(event.request_id, None)
        if activity:
            activity.set_property("db.duration_ms", event.duration_micros / 1000)
            activity.set_property("exception.message", str(event.failure))
            activity.set_status(StatusCode.ERROR, str(event.failure))
            activity.stop()



from typing import Dict, Mapping, Optional
from opentelemetry import trace, baggage
from opentelemetry.trace import (
    get_current_span,
    Span,
    Status,
    StatusCode
)
from opentelemetry.context import get_current, attach, detach

_tracer = trace.get_tracer("blocks.activity")


class Activity:
    def __init__(self, name: str):
        self._context = get_current()
        self._parent_span = get_current_span()
        self._root_attributes = self._find_root_attributes(self._parent_span)

        self._span = _tracer.start_span(name, context=self._context)

        if self._span.is_recording() and self._root_attributes:
            for k, v in self._root_attributes.items():
                self._span.set_attribute(f"baggage.{k}", v)

        self._span_context = trace.set_span_in_context(self._span, self._context)
        self._token = attach(self._span_context)
        
    def _find_root_attributes(self, span: Optional[Span]) -> Dict[str, object]:
        """Find root attributes in the span or its parent."""
        if span is None or not hasattr(span, 'attributes'):
            return {}

        root_attrs = {}
        for k, v in span.attributes.items():
            if k.startswith("baggage."):
                root_attrs[k[8:]] = v

        if root_attrs:
            return root_attrs

        return dict(span.attributes)

    def get_root_attribute(self, key: str) -> Optional[object]:
        """Get a single attribute from the root span attributes."""
        if self._root_attributes and key in self._root_attributes:
            return self._root_attributes[key]
        return None

    def get_all_root_attributes(self) -> Dict[str, object]:
        """Get all attributes from the root span."""
        return self._root_attributes or {}

    def set_property(self, key: str, value):
        if self._span.is_recording():
            self._span.set_attribute(key, value)

    def set_properties(self, props: dict):
        if self._span.is_recording():
            for k, v in props.items():
                self._span.set_attribute(k, v)

    def set_status(self, status_code: StatusCode, description: str = ""):
        self._span.set_status(Status(status_code, description))

    def stop(self):
        self._span.end()
        detach(self._token)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            self._span.record_exception(exc_val)
            self.set_status(StatusCode.ERROR, str(exc_val))
        self.stop()

    @staticmethod
    def current() -> Span:
        return get_current_span()

    @staticmethod
    def start(name: str) -> "Activity":
        return Activity(name)

    @staticmethod
    def get_trace_id() -> str:
        span = Activity.current()
        return format(span.get_span_context().trace_id, "032x") if span else ""

    @staticmethod
    def get_span_id() -> str:
        span = Activity.current()
        return format(span.get_span_context().span_id, "016x") if span else ""

    @staticmethod
    def set_current_property(key: str, value):
        span = Activity.current()
        if span and span.is_recording():
            span.set_attribute(key, value)

    @staticmethod
    def set_current_properties(props: dict):
        span = Activity.current()
        if span and span.is_recording():
            for k, v in props.items():
                span.set_attribute(k, v)


from typing import Any


class CacheProvider:
    """
    Singleton class for managing a global cache client.
    This client will be shared throughout the entire application lifecycle.
    """
    _cache_client: Any = None

    @staticmethod
    def set_client(cache_client: Any) -> None:
        """
        Set the global cache client implementation.

        Args:
            cache_client: An instance of the cache client to use.
        """
        CacheProvider._cache_client = cache_client

    @staticmethod
    def get_client() -> Any:
        """
        Get the global cache client.

        Returns:
            The configured cache client.

        Raises:
            RuntimeError: If the cache client is not initialized.
        """
        if CacheProvider._cache_client is None:
            raise RuntimeError("Cache client not initialized")
        return CacheProvider._cache_client

    @staticmethod
    def clear() -> None:
        """
        Clear the global cache client.
        """
        CacheProvider._cache_client = None



import json
import logging
import threading
from dataclasses import asdict, is_dataclass
from typing import Any, Dict, Optional

from asyncio import Lock
from collections import defaultdict
from datetime import datetime
from azure.servicebus.aio import ServiceBusClient, ServiceBusSender
from azure.servicebus import ServiceBusMessage

from blocks_genesis.auth.blocks_context import BlocksContextManager
from blocks_genesis.core.secret_loader import get_blocks_secret
from blocks_genesis.lmt.activity import Activity
from blocks_genesis.message.consumer_message import ConsumerMessage
from blocks_genesis.message.event_message import EventMessage
from blocks_genesis.message.message_client import MessageClient
from blocks_genesis.message.message_configuration import MessageConfiguration

logger = logging.getLogger(__name__)

class DateTimeEncoder(json.JSONEncoder):
    def default(self, obj: Any) -> Any:
        if isinstance(obj, datetime):
            return obj.isoformat()
        return super().default(obj)

class AzureMessageClient(MessageClient):
    _instance: Optional['AzureMessageClient'] = None
    _singleton_lock = threading.Lock()

    def __init__(self, message_config: MessageConfiguration):
        self._message_config = message_config
        self._client = ServiceBusClient.from_connection_string(message_config.connection or get_blocks_secret().MessageConnectionString)
        self._senders: Dict[str, ServiceBusSender] = {}
        self._sender_locks: Dict[str, Lock] = defaultdict(Lock)
        self._initialize_senders()

    @classmethod
    def initialize(cls, message_config: MessageConfiguration):
        with cls._singleton_lock:
            if cls._instance is None:
                cls._instance = cls(message_config)
                logger.info("✅ AzureMessageClient singleton initialized.")

    @classmethod
    def get_instance(cls) -> 'AzureMessageClient':
        if cls._instance is None:
            raise Exception("AzureMessageClient not initialized. Call `initialize()` first.")
        return cls._instance

    def _initialize_senders(self):
        queues = self._message_config.azure_service_bus_configuration.queues or []
        topics = self._message_config.azure_service_bus_configuration.topics or []
        logger.info(f"Initializing Azure Service Bus senders for queues: {queues} and topics: {topics}")

        for name in queues + topics:
            self._senders[name] = (
                self._client.get_queue_sender(queue_name=name)
                if name in queues
                else self._client.get_topic_sender(topic_name=name)
            )

    async def _get_sender(self, name: str) -> ServiceBusSender:
        if name in self._senders:
            return self._senders[name]

        async with self._sender_locks[name]:
            if name not in self._senders:
                self._senders[name] = self._client.get_topic_sender(topic_name=name)
            return self._senders[name]


    async def _send_to_azure_bus_async(self, consumer_message: ConsumerMessage, is_topic: bool = False):
        security_context = BlocksContextManager.get_context()

        with Activity("messaging.azure.servicebus.send") as activity:
            activity.set_properties({
                "messaging.system": "azure.servicebus",
                "messaging.destination": consumer_message.consumer_name,
                "messaging.destination_kind": "topic" if is_topic else "queue",
                "messaging.operation": "send",
                "messaging.message_type": type(consumer_message.payload).__name__,
                "baggage.TenantId": BlocksContextManager.get_context().tenant_id
            })

            sender = await self._get_sender(consumer_message.consumer_name)

            payload_dict = self._serialize_payload(consumer_message.payload)

            message_body = EventMessage(
                body=json.dumps(payload_dict),
                type=type(consumer_message.payload).__name__
            )

            sb_message = ServiceBusMessage(
                body=json.dumps(message_body.__dict__),
                application_properties={
                    "TenantId": security_context.tenant_id if security_context else None,
                    "TraceId": activity.get_trace_id(),
                    "SpanId": activity.get_span_id(),
                    "SecurityContext": consumer_message.context or json.dumps(
                        security_context.__dict__ if security_context else {}, 
                        cls=DateTimeEncoder
                    ),
                    "Baggage": json.dumps(activity.get_all_root_attributes())
                }
            )

            await sender.send_messages(sb_message)

    def _serialize_payload(self, payload):
        if is_dataclass(payload):
            return asdict(payload)
        elif isinstance(payload, dict):
            return payload
        elif isinstance(payload, str):
            return {"message": payload}
        else:
            raise TypeError(f"Unsupported payload type: {type(payload)}")

    async def send_to_consumer_async(self, consumer_message: ConsumerMessage):
        await self._send_to_azure_bus_async(consumer_message)

    async def send_to_mass_consumer_async(self, consumer_message: ConsumerMessage):
        await self._send_to_azure_bus_async(consumer_message, is_topic=True)

    async def close(self):
        await self._client.close()



from abc import ABC, abstractmethod
from pymongo.collection import Collection
from pymongo.database import Database
from typing import Optional


class DbContextProvider(ABC):
    @abstractmethod
    async def get_database(self, tenant_id: Optional[str] = None) -> Optional[Database]:
        pass

    @abstractmethod
    async def get_collection(self, collection_name: str, tenant_id: Optional[str] = None) -> Collection:
        pass



from typing import Optional
from azure.servicebus.management import ServiceBusAdministrationClient
from blocks_genesis.message.message_configuration import MessageConfiguration


class ConfigAzureServiceBus:
    _admin_client: Optional[ServiceBusAdministrationClient] = None
    _message_config: Optional[MessageConfiguration] = None

    @classmethod
    def configure_queue_and_topic(cls, message_config: MessageConfiguration):
        if not message_config.connection:
            print("Error: Missing Service Bus connection string.")
            return

        try:
            cls._admin_client = ServiceBusAdministrationClient.from_connection_string(message_config.connection)
            cls._message_config = message_config

            cls._create_queues()
            cls._create_topics_and_subscriptions()

        except Exception as ex:
            print(f"Exception during Service Bus configuration: {ex}")
            raise

    @classmethod
    def _create_queues(cls):
        queues = cls._message_config.azure_service_bus_configuration.queues or []

        for queue_name in queues:
            if cls._check_queue_exists(queue_name):
                print(f"Queue '{queue_name}' already exists. Skipping creation.")
                continue

            cls._admin_client.create_queue(
                queue_name,
                max_size_in_megabytes=cls._message_config.azure_service_bus_configuration.queue_max_size_in_megabytes,
                max_delivery_count=cls._message_config.azure_service_bus_configuration.queue_max_delivery_count,
                default_message_time_to_live=cls._message_config.azure_service_bus_configuration.queue_default_message_time_to_live
            )
            print(f"✅ Queue created: {queue_name}")

    @classmethod
    def _check_queue_exists(cls, queue_name: str) -> bool:
        try:
            cls._admin_client.get_queue(queue_name)
            return True
        except Exception:
            return False

    @classmethod
    def _create_topics_and_subscriptions(cls):
        topics = cls._message_config.azure_service_bus_configuration.topics or []

        for topic_name in topics:
            if cls._check_topic_exists(topic_name):
                print(f"Topic '{topic_name}' already exists. Skipping creation.")
            else:
                cls._admin_client.create_topic(
                    topic_name,
                    max_size_in_megabytes=cls._message_config.azure_service_bus_configuration.topic_max_size_in_megabytes,
                    default_message_time_to_live=cls._message_config.azure_service_bus_configuration.topic_default_message_time_to_live
                )
                print(f"✅ Topic created: {topic_name}")

            cls._create_subscription(topic_name)

    @classmethod
    def _check_topic_exists(cls, topic_name: str) -> bool:
        try:
            cls._admin_client.get_topic(topic_name)
            return True
        except Exception:
            return False

    @classmethod
    def _create_subscription(cls, topic_name: str):
        subscription_name = cls._message_config.get_subscription_name(topic_name)
        if cls._check_subscription_exists(topic_name, subscription_name):
            print(f"Subscription '{subscription_name}' for topic '{topic_name}' already exists. Skipping.")
            return

        cls._admin_client.create_subscription(
            topic_name,
            subscription_name,
            max_delivery_count=cls._message_config.azure_service_bus_configuration.topic_subscription_max_delivery_count,
            default_message_time_to_live=cls._message_config.azure_service_bus_configuration.topic_subscription_default_message_time_to_live
        )
        print(f"✅ Subscription '{subscription_name}' created for topic '{topic_name}'")

    @classmethod
    def _check_subscription_exists(cls, topic_name: str, subscription_name: str) -> bool:
        try:
            cls._admin_client.get_subscription(topic_name, subscription_name)
            return True
        except Exception:
            return False



import asyncio
import logging
import re
from typing import Dict, Optional, Tuple
from motor.motor_asyncio import AsyncIOMotorClient

from blocks_genesis.cache import CacheClient
from blocks_genesis.cache.cache_provider import CacheProvider
from blocks_genesis.core.secret_loader import get_blocks_secret
from blocks_genesis.tenant.tenant import Tenant

_logger = logging.getLogger(__name__)

class TenantService:
    """Manages tenant configuration with caching and real-time updates"""

    def __init__(self):
        self._blocks_secret = get_blocks_secret()
        self.cache: CacheClient = CacheProvider.get_client()
        if not self.cache:
            raise RuntimeError("Cache client not initialized")

        self.client = AsyncIOMotorClient(self._blocks_secret.DatabaseConnectionString)
        self.database = self.client[self._blocks_secret.RootDatabaseName]

        self._tenant_cache: Dict[str, Tenant] = {}
        self._update_channel = "tenant::updates"
        self._collection_name = "Tenants"

        self._initialized = False
        self._initialize_lock = asyncio.Lock()

    async def initialize(self):
        """Explicit initializer for async setup"""
        async with self._initialize_lock:
            if self._initialized:
                return

            await self._load_tenants()
            asyncio.create_task(self._subscribe_to_updates())
            self._initialized = True
            _logger.info("TenantService initialized successfully")

    async def get_tenant(self, tenant_id: str) -> Optional[Tenant]:
        if not tenant_id:
            return None

        tenant = self._tenant_cache.get(tenant_id)
        if tenant:
            return tenant

        tenant = await self._load_tenant_from_db(tenant_id)
        if tenant:
            self._tenant_cache[tenant.tenant_id] = tenant
        return tenant

    async def get_tenant_by_domain(self, domain: str) -> Optional[Tenant]:
        if not domain:
            return None
        try:
            tenant_dict = await self.database[self._collection_name].find_one({
                "$or": [
                    {"ApplicationDomain": domain},
                    {"ApplicationDomain": {"$regex": re.compile(domain)}},
                    {"AllowedDomains": {"$in": [domain]}}
                ]
            })
            if tenant_dict:
                tenant = Tenant(**tenant_dict)
                self._tenant_cache[tenant.tenant_id] = tenant
                return tenant
        except Exception as e:
            _logger.exception(f"Error getting tenant by domain {domain}: {e}")
        return None

    async def get_db_connection(self, tenant_id: str) -> Tuple[Optional[str], Optional[str]]:
        tenant = await self.get_tenant(tenant_id)
        if tenant:
            return tenant.db_name, tenant.db_connection_string
        return None, None

    async def _load_tenants(self):
        try:
            cursor = self.database[self._collection_name].find({})
            self._tenant_cache.clear()
            async for tenant_dict in cursor:
                tenant = Tenant(**tenant_dict)
                self._tenant_cache[tenant.tenant_id] = tenant
            _logger.info(f"Loaded {len(self._tenant_cache)} tenants into cache")
        except Exception as e:
            _logger.exception(f"Failed to load tenants: {e}")

    async def _load_tenant_from_db(self, tenant_id: str) -> Optional[Tenant]:
        try:
            tenant_dict = await self.database[self._collection_name].find_one({
                "$or": [
                    {"_id": tenant_id},
                    {"TenantId": tenant_id}
                ]
            })
            if tenant_dict:
                return Tenant(**tenant_dict)
        except Exception as e:
            _logger.exception(f"Error loading tenant {tenant_id}: {e}")
        return None

    async def _subscribe_to_updates(self):
        try:
            await self.cache.subscribe_async(
                self._update_channel,
                self._handle_update
            )
            _logger.info("Subscribed to tenant updates")
        except Exception as e:
            _logger.exception(f"Failed to subscribe to updates: {e}")

    async def _handle_update(self, channel: str, message: str):
        try:
            _logger.info(f"Received tenant update: {message}")
            await self._load_tenants()
        except Exception as e:
            _logger.exception(f"Error handling update: {e}")


# Global tenant service singleton instance
_tenant_service: Optional[TenantService] = None

def get_tenant_service() -> TenantService:
    if _tenant_service is None:
        raise RuntimeError("TenantService not initialized. Call initialize_tenant_service() first.")
    return _tenant_service

async def initialize_tenant_service() -> TenantService:
    global _tenant_service
    if _tenant_service is None:
        _tenant_service = TenantService()
    await _tenant_service.initialize()
    return _tenant_service



from contextvars import ContextVar
from datetime import datetime
from typing import ClassVar, List, Optional, Dict, Any
from pydantic import BaseModel, Field
import threading

class BlocksContext(BaseModel):
   # JWT Standard Claims
    ISSUER_CLAIM: ClassVar[str] = "iss"
    AUDIENCES_CLAIM: ClassVar[str] = "aud"
    ISSUED_AT_TIME_CLAIM: ClassVar[str] = "iat"
    NOT_BEFORE_THAT_CLAIM: ClassVar[str] = "nbf"
    EXPIRE_ON_CLAIM: ClassVar[str] = "exp"
    SUBJECT_CLAIM: ClassVar[str] = "sub"
    
    # Custom Claims
    TENANT_ID_CLAIM: ClassVar[str] = "tenant_id"
    ROLES_CLAIM: ClassVar[str] = "roles"
    USER_ID_CLAIM: ClassVar[str] = "user_id"
    REQUEST_URI_CLAIM: ClassVar[str] = "request_uri"
    TOKEN_CLAIM: ClassVar[str] = "oauth"
    PERMISSION_CLAIM: ClassVar[str] = "permissions"
    ORGANIZATION_ID_CLAIM: ClassVar[str] = "org_id"
    EMAIL_CLAIM: ClassVar[str] = "email"
    USER_NAME_CLAIM: ClassVar[str] = "user_name"
    DISPLAY_NAME_CLAIM: ClassVar[str] = "name"
    PHONE_NUMBER_CLAIM: ClassVar[str] = "phone"
    
    # Properties
    tenant_id: str = ""
    roles: List[str] = Field(default_factory=list)
    user_id: str = ""
    expire_on: Optional[datetime] = None
    request_uri: str = ""
    oauth_token: str = ""
    organization_id: str = ""
    is_authenticated: bool = False
    email: str = ""
    permissions: List[str] = Field(default_factory=list)
    user_name: str = ""
    phone_number: str = ""
    display_name: str = ""
    
    class Config:
        arbitrary_types_allowed = True

# Context variables for async context management
_context_var: ContextVar[Optional[BlocksContext]] = ContextVar('blocks_context', default=None)
_test_mode = threading.local()

class BlocksContextManager:
    """Manages BlocksContext instances and provides utility methods"""
    
    @staticmethod
    def get_test_mode() -> bool:
        """Get test mode status (thread-safe)"""
        return getattr(_test_mode, 'value', False)
    
    @staticmethod
    def set_test_mode(value: bool) -> None:
        """Set test mode status (thread-safe)"""
        _test_mode.value = value
    
    @staticmethod
    def create_from_jwt_claims(claims: Dict[str, Any]) -> BlocksContext:
        """Create BlocksContext from JWT claims dictionary"""
        
        def get_claim_value(claim_name: str, default: Any = "") -> Any:
            return claims.get(claim_name, default)
        
        def get_claim_list(claim_name: str) -> List[str]:
            value = claims.get(claim_name, [])
            if isinstance(value, str):
                return [value]
            return value if isinstance(value, list) else []
        
        expire_on = None
        if exp_claim := claims.get(BlocksContext.EXPIRE_ON_CLAIM):
            try:
                if isinstance(exp_claim, (int, float)):
                    expire_on = datetime.fromtimestamp(exp_claim)
                elif isinstance(exp_claim, str):
                    expire_on = datetime.fromisoformat(exp_claim.replace('Z', '+00:00'))
            except (ValueError, TypeError):
                expire_on = None
        
        return BlocksContext(
            tenant_id=get_claim_value(BlocksContext.TENANT_ID_CLAIM),
            roles=get_claim_list(BlocksContext.ROLES_CLAIM),
            user_id=get_claim_value(BlocksContext.USER_ID_CLAIM),
            is_authenticated=True,
            request_uri=get_claim_value(BlocksContext.REQUEST_URI_CLAIM),
            organization_id=get_claim_value(BlocksContext.ORGANIZATION_ID_CLAIM),
            expire_on=expire_on,
            email=get_claim_value(BlocksContext.EMAIL_CLAIM),
            permissions=get_claim_list(BlocksContext.PERMISSION_CLAIM),
            user_name=get_claim_value(BlocksContext.USER_NAME_CLAIM),
            phone_number=get_claim_value(BlocksContext.PHONE_NUMBER_CLAIM),
            display_name=get_claim_value(BlocksContext.DISPLAY_NAME_CLAIM),
            oauth_token=get_claim_value(BlocksContext.TOKEN_CLAIM)
        )
    
    @staticmethod
    def create(
        tenant_id: Optional[str] = None,
        roles: Optional[List[str]] = None,
        user_id: Optional[str] = None,
        is_authenticated: bool = False,
        request_uri: Optional[str] = None,
        organization_id: Optional[str] = None,
        expire_on: Optional[datetime] = None,
        email: Optional[str] = None,
        permissions: Optional[List[str]] = None,
        user_name: Optional[str] = None,
        phone_number: Optional[str] = None,
        display_name: Optional[str] = None,
        oauth_token: Optional[str] = None
    ) -> BlocksContext:
        """Create BlocksContext from individual parameters"""
        return BlocksContext(
            tenant_id=tenant_id or "",
            roles=roles or [],
            user_id=user_id or "",
            is_authenticated=is_authenticated,
            request_uri=request_uri or "",
            organization_id=organization_id or "",
            expire_on=expire_on,
            email=email or "",
            permissions=permissions or [],
            user_name=user_name or "",
            phone_number=phone_number or "",
            display_name=display_name or "",
            oauth_token=oauth_token or ""
        )
    
    @staticmethod
    def get_context(test_value: Optional[BlocksContext] = None) -> Optional[BlocksContext]:
        """Get the current BlocksContext"""
        try:
            # For testing scenarios
            if BlocksContextManager.get_test_mode():
                return test_value or _context_var.get()
            
            return _context_var.get()
        except Exception:
            return None
    
    @staticmethod
    def set_context(context: Optional[BlocksContext]) -> None:
        """Set the context in ContextVar storage"""
        _context_var.set(context)
    
    @staticmethod
    def clear_context() -> None:
        """Clear the current context"""
        _context_var.set(None)
    
    




import logging
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.requests import Request
from starlette.responses import JSONResponse

from blocks_genesis.lmt.activity import Activity

_logger = logging.getLogger(__name__)

class GlobalExceptionHandlerMiddleware(BaseHTTPMiddleware):
    JSON_CONTENT_TYPE = "application/json"
    EMPTY_JSON_BODY = "Empty"
    MAX_PAYLOAD_LENGTH = 1000


    async def dispatch(self, request: Request, call_next):
        try:
            return await call_next(request)
        except Exception as exc:
            return await self.handle_exception(request, exc)

    async def handle_exception(self, request: Request, exception: Exception):
        payload = self.EMPTY_JSON_BODY

        if request.headers.get("content-type", "").startswith(self.JSON_CONTENT_TYPE):
            try:
                body = await request.body()
                payload = body.decode("utf-8").strip()
                if not payload:
                    payload = self.EMPTY_JSON_BODY
                elif len(payload) > self.MAX_PAYLOAD_LENGTH:
                    payload = payload[:self.MAX_PAYLOAD_LENGTH] + "... [truncated]"
            except Exception:
                payload = "[Unable to read request body]"


        trace_id = Activity.get_trace_id()

        url = str(request.url)
        method = request.method
        message = (
            f"Unhandled exception thrown on request Trace: [{trace_id}] "
            f"Method: [{method}] {url} : {str(exception)}.\n"
            f"Payload: {payload}"
        )
        _logger.exception(message)

        # Return standardized JSON error
        response = {
            "Message": "An error occurred while processing your request.",
            "TraceId": trace_id
        }

        return JSONResponse(content=response, status_code=500)



from datetime import datetime
from typing import List, Optional
from pydantic import BaseModel, Field
from enum import IntEnum

from blocks_genesis.entities.base_entity import BaseEntity


# CertificateStorageType enum
class CertificateStorageType(IntEnum):
    AZURE = 1
    FILESYSTEM = 2
    MONGODB = 3

# JwtTokenParameters class
class JwtTokenParameters(BaseModel):
    issuer: str = Field(alias="Issuer")
    subject: Optional[str] = Field(alias="Subject", default=None)
    audiences: List[str] = Field(alias="Audiences")
    public_certificate_path: Optional[str] = Field(alias="PublicCertificatePath", default=None)
    public_certificate_password: str = Field(alias="PublicCertificatePassword")
    private_certificate_password: str = Field(alias="PrivateCertificatePassword")
    certificate_storage_type: CertificateStorageType = Field(alias="CertificateStorageType", default=CertificateStorageType.AZURE)
    certificate_valid_for_number_of_days: int = Field(alias="CertificateValidForNumberOfDays")
    issue_date: datetime = Field(alias="IssueDate")

    class Config:
        extra = "ignore"
        validate_by_name = True
        use_enum_values = True

# Tenant class for MongoDB document mapping
class Tenant(BaseEntity):
    tenant_id: str = Field(alias="TenantId")
    is_accept_blocks_terms: bool = Field(alias="IsAcceptBlocksTerms")
    is_use_blocks_exclusively: bool = Field(alias="IsUseBlocksExclusively")
    is_production: bool = Field(alias="IsProduction")
    name: Optional[str] = Field(alias="Name", default=None)
    db_name: str = Field(alias="DBName")
    application_domain: str = Field(alias="ApplicationDomain")
    allowed_domains: List[str] = Field(default_factory=list, alias="AllowedDomains")
    cookie_domain: str = Field(alias="CookieDomain")
    is_disabled: bool = Field(alias="IsDisabled")
    db_connection_string: str = Field(alias="DbConnectionString")
    tenant_salt: str = Field(alias="TenantSalt")
    jwt_token_parameters: JwtTokenParameters = Field(alias="JwtTokenParameters")
    is_root_tenant: bool = Field(alias="IsRootTenant")
    is_cookie_enable: bool = Field(alias="IsCookieEnable")
    is_domain_verified: bool = Field(alias="IsDomainVerified")

    class Config:
        extra = "ignore"
        validate_by_name = True


from fastapi import Depends, Request, HTTPException
from pymongo.collection import Collection
from datetime import datetime, timezone
import base64
import aiohttp
import asyncio
import jwt
from jwt import InvalidTokenError

from cryptography.hazmat.primitives.serialization import pkcs12
from cryptography.hazmat.primitives import serialization

from blocks_genesis.auth.blocks_context import BlocksContext, BlocksContextManager
from blocks_genesis.cache import CacheClient
from blocks_genesis.cache.cache_provider import CacheProvider
from blocks_genesis.database.db_context import DbContext
from blocks_genesis.tenant.tenant import Tenant
from blocks_genesis.tenant.tenant_service import TenantService, get_tenant_service



async def fetch_cert_bytes(cert_url: str) -> bytes:
    if cert_url.startswith("http"):
        async with aiohttp.ClientSession() as session:
            async with session.get(cert_url) as resp:
                resp.raise_for_status()
                return await resp.read()
    else:
        loop = asyncio.get_running_loop()
        try:
            with open(cert_url, "rb") as f:
                return await loop.run_in_executor(None, f.read)
        except Exception as e:
            raise RuntimeError(f"Error reading cert file {cert_url}: {e}")

async def get_tenant_cert(cache_client: CacheClient, tenant: Tenant, tenant_id: str) -> bytes:
    key = f"tetocertpublic::{tenant_id}"
    cert_bytes = cache_client.get_string_value(key)
    if cert_bytes is None:
        cert_bytes = await fetch_cert_bytes(tenant.jwt_token_parameters.public_certificate_path)
        now = datetime.now(timezone.utc)
        issue_date = tenant.jwt_token_parameters.issue_date
        if issue_date.tzinfo is None:
            issue_date = issue_date.replace(tzinfo=timezone.utc)
        days_remaining = (
            tenant.jwt_token_parameters.certificate_valid_for_number_of_days
            - (now - issue_date).days
            - 1
        )
        ttl = max(60, days_remaining  * 24 * 60 * 60)  # Ensure at least 60 seconds TTL
        if ttl > 0:
            cached_value = base64.b64encode(cert_bytes).decode("utf-8")
            await cache_client.add_string_value(key, cached_value, ex=int(ttl))
    return cert_bytes


async def authenticate(request: Request, tenant_service: TenantService, cache_client: CacheClient):
    header = request.headers.get("Authorization")
    if header and any(header.startswith(prefix) for prefix in ["bearer ", "Bearer "]):
        token = header.split(" ", 1)[1].strip()
    else:
        token = request.cookies.get("access_token", "")
        
    if not token:
        raise HTTPException(401, "Token missing")

    tenant_id = BlocksContextManager.get_context().tenant_id if BlocksContextManager.get_context() else None
    if not tenant_id:
        raise HTTPException(401, "Tenant ID missing")

    tenant = await tenant_service.get_tenant(tenant_id)
    cert_bytes = await get_tenant_cert(cache_client, tenant, tenant_id)
    
    cert = create_certificate(cert_bytes, tenant.jwt_token_parameters.public_certificate_password)
    if not cert:
        raise HTTPException(500, "Failed to load certificate")
    
    public_key = cert.public_key()
    public_key_pem = public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo
    )
    public_key_str = public_key_pem.decode('utf-8')
    
    try:
        payload = jwt.decode(
            jwt=token,
            key=public_key_str,
            algorithms=["RS256"],
            issuer=tenant.jwt_token_parameters.issuer,
            audience=tenant.jwt_token_parameters.audiences,
            options={
                "verify_signature": True,  
                "verify_exp": True,        
                "verify_iss": True,        
                "verify_aud": True, 
                "verify_iat": True,
                "verify_nbf": True,      
                "require":["exp", "iat", "iss", "aud", "nbf"]
            },
            leeway=0 
        )
        extended_payload = dict(payload)
        extended_payload[BlocksContext.REQUEST_URI_CLAIM] = str(request.url)
        extended_payload[BlocksContext.TOKEN_CLAIM] = token

        BlocksContextManager.create_from_jwt_claims(extended_payload)
        return extended_payload
    except InvalidTokenError as e:
        print(f"JWT verification failed: {e}")
        raise HTTPException(401, f"Invalid token: {e}")

    

def create_certificate(certificate_data: bytes, password: str = None):
    """Load public certificate from PFX data"""
    try:
        password_bytes = password.encode('utf-8') if password else None
        certificate = pkcs12.load_pkcs12(certificate_data, password_bytes)
        return certificate.additional_certs[0].certificate
    except Exception as e:
        print(f"Failed to create certificate: {e}")
        return None

def authorize(bypass_authorization: bool = False):
    async def dependency(request: Request):
        tenant_service = get_tenant_service()
        cache_client = CacheProvider.get_client()
        db_context = DbContext.get_provider()

        await authenticate(request, tenant_service, cache_client)
        context = BlocksContextManager.get_context()
        if not context:
            raise HTTPException(401, "Missing context")

        if bypass_authorization:
            return

        roles = context.roles or []
        permissions = context.permissions or []

        # Parse controller and action from URL path
        path_parts = request.url.path.strip("/").split("/")
        if len(path_parts) >= 4:
            controller = path_parts[2]
            action = path_parts[3]
        elif len(path_parts) >= 2:
            controller = path_parts[0]
            action = path_parts[1]
        else:
            raise HTTPException(400, "Invalid URL format.")

        resource = f"{context.service_name}::{controller}::{action}".lower()

        collection: Collection = await db_context.get_collection("Permissions", tenant_id=context.tenant_id)

        query = {
            "Type": 1,
            "Resource": resource,
            "$or": [
                {"Roles": {"$in": roles}},
                {"Name": {"$in": permissions}}
            ]
        }

        count = await collection.count_documents(query)
        if count < 1:
            raise HTTPException(403, "Insufficient permissions")

    return Depends(dependency)



# interfaces/cache_client.py
from abc import ABC, abstractmethod
from typing import Dict, Optional, Callable, Any
import redis


class CacheClient(ABC):
    """Abstract base class for cache client implementations"""
    
    @abstractmethod
    def cache_database(self) -> redis.Redis:
        """Get the cache database instance"""
        pass
    
    # Synchronous Methods
    @abstractmethod
    def key_exists(self, key: str) -> bool:
        """Check if key exists in cache"""
        pass
    
    @abstractmethod
    def add_string_value(self, key: str, value: str, key_life_span: Optional[int] = None) -> bool:
        """Add string value to cache with optional TTL"""
        pass
    
    @abstractmethod
    def get_string_value(self, key: str) -> Optional[str]:
        """Get string value from cache"""
        pass
    
    @abstractmethod
    def remove_key(self, key: str) -> bool:
        """Remove key from cache"""
        pass
    
    @abstractmethod
    def add_hash_value(self, key: str, value: Dict[str, Any], key_life_span: Optional[int] = None) -> bool:
        """Add hash value to cache with optional TTL"""
        pass
    
    @abstractmethod
    def get_hash_value(self, key: str) -> Dict[str, Any]:
        """Get hash value from cache"""
        pass
    
    # Asynchronous Methods
    @abstractmethod
    async def key_exists_async(self, key: str) -> bool:
        """Check if key exists in cache (async)"""
        pass
    
    @abstractmethod
    async def add_string_value_async(self, key: str, value: str, key_life_span: Optional[int] = None) -> bool:
        """Add string value to cache with optional TTL (async)"""
        pass
    
    @abstractmethod
    async def get_string_value_async(self, key: str) -> Optional[str]:
        """Get string value from cache (async)"""
        pass
    
    @abstractmethod
    async def remove_key_async(self, key: str) -> bool:
        """Remove key from cache (async)"""
        pass
    
    @abstractmethod
    async def add_hash_value_async(self, key: str, value: Dict[str, Any], key_life_span: Optional[int] = None) -> bool:
        """Add hash value to cache with optional TTL (async)"""
        pass
    
    @abstractmethod
    async def get_hash_value_async(self, key: str) -> Dict[str, Any]:
        """Get hash value from cache (async)"""
        pass
    
    # Pub/Sub Methods
    @abstractmethod
    async def publish_async(self, channel: str, message: str) -> int:
        """Publish message to channel"""
        pass
    
    @abstractmethod
    async def subscribe_async(self, channel: str, handler: Callable[[str, str], None]) -> None:
        """Subscribe to channel with message handler"""
        pass
    
    @abstractmethod
    async def unsubscribe_async(self, channel: str) -> None:
        """Unsubscribe from channel"""
        pass
    
    # Dispose Methods
    @abstractmethod
    def dispose(self) -> None:
        """Dispose resources synchronously"""
        pass
    
    @abstractmethod
    async def dispose_async(self) -> None:
        """Dispose resources asynchronously"""
        pass
    
    # Context Manager Support
    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.dispose()
    
    async def __aenter__(self):
        return self
    
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.dispose_async()


from abc import ABC, abstractmethod

from blocks_genesis.message.consumer_message import ConsumerMessage


class MessageClient(ABC):
    @abstractmethod
    async def send_to_consumer_async(self, consumer_message: ConsumerMessage) -> None:
        pass

    @abstractmethod
    async def send_to_mass_consumer_async(self, consumer_message: ConsumerMessage) -> None:
        pass



from pydantic import BaseModel


class BlocksSecret(BaseModel):
    CacheConnectionString: str = ""
    MessageConnectionString: str = ""
    LogConnectionString: str = ""
    MetricConnectionString: str = ""
    TraceConnectionString: str = ""
    LogDatabaseName: str = ""
    MetricDatabaseName: str = ""
    TraceDatabaseName: str = ""
    ServiceName: str = ""
    DatabaseConnectionString: str = ""
    RootDatabaseName: str = ""




import logging
from fastapi import FastAPI, logger
from fastapi.middleware.httpsredirect import HTTPSRedirectMiddleware
from fastapi.middleware.gzip import GZipMiddleware
from starlette.middleware.cors import CORSMiddleware
from opentelemetry.instrumentation.fastapi import FastAPIInstrumentor
from blocks_genesis.cache.cache_provider import CacheProvider
from blocks_genesis.cache.redis_client import RedisClient
from blocks_genesis.core.secret_loader import SecretLoader
from blocks_genesis.database.db_context import DbContext
from blocks_genesis.database.mongo_context import MongoDbContextProvider
from blocks_genesis.lmt.log_config import configure_logger
from blocks_genesis.lmt.mongo_log_exporter import MongoHandler
from blocks_genesis.lmt.tracing import configure_tracing
from blocks_genesis.message.azure.azure_message_client import AzureMessageClient
from blocks_genesis.message.message_configuration import MessageConfiguration
from blocks_genesis.middlewares.global_exception_middleware import GlobalExceptionHandlerMiddleware
from blocks_genesis.middlewares.tenant_middleware import TenantValidationMiddleware
from blocks_genesis.tenant.tenant_service import initialize_tenant_service

logger = logging.getLogger(__name__)

async def configure_lifespan(name: str, message_config: MessageConfiguration):
    logger.info("🚀 Initializing services...")
    logger.info("🔐 Loading secrets before app creation...")
    secret_loader = SecretLoader(name)
    await secret_loader.load_secrets()
    logger.info("✅ Secrets loaded successfully!")
    
    configure_logger()
    logger.info("Logger started")

    # Enable tracing after secrets are loaded
    configure_tracing()
    logger.info("🔍 Tracing enabled successfully!")

    CacheProvider.set_client(RedisClient())
    await initialize_tenant_service()
    DbContext.set_provider(MongoDbContextProvider())
    
    AzureMessageClient.initialize(message_config)
    
    
async def close_lifespan():
    logger.info("🛑 Shutting down services...")
    
    await AzureMessageClient.get_instance().close()
    # Shutdown logic
    if hasattr(MongoHandler, '_mongo_logger') and MongoHandler._mongo_logger:
        MongoHandler._mongo_logger.stop()
        
def configure_middlewares(app: FastAPI, is_local: False):
    if not is_local:
        app.add_middleware(HTTPSRedirectMiddleware)
        
    app.add_middleware(GZipMiddleware)
    app.add_middleware(TenantValidationMiddleware)
    app.add_middleware(GlobalExceptionHandlerMiddleware)
    FastAPIInstrumentor.instrument_app(app)  ### Instrument FastAPI for OpenTelemetry
    app.add_middleware(
        CORSMiddleware,
        allow_origins=["*"],
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )
    
    @app.get("/ping")
    async def health():
        return {
            "status": "healthy",
            "secrets_status": "loaded" ,
        }


# mongo_trace_exporter.py

from datetime import datetime
import threading
from queue import Queue, Empty
from pymongo import MongoClient, errors
from opentelemetry.sdk.trace.export import SpanExporter, SpanExportResult
from blocks_genesis.core.secret_loader import get_blocks_secret
import time


class MongoDBTraceExporter(SpanExporter):
    def __init__(self, flush_interval: float = 3.0, batch_size: int = 1000, queue_size: int = 10000):
        self._blocks_secret = get_blocks_secret()
        self._service_name = self._blocks_secret.ServiceName
        self._client = MongoClient(self._blocks_secret.TraceConnectionString)
        self._db = self._client[self._blocks_secret.TraceDatabaseName]
        
        self._queue = Queue(maxsize=queue_size)
        self._batch_size = batch_size
        self._flush_interval = flush_interval
        
        self._stop_event = threading.Event()
        self._worker_thread = threading.Thread(target=self._run, daemon=True)
        self._worker_thread.start()

    def _extract_baggage_from_span(self, span) -> dict:
        baggage_items = {}

        for key in span.attributes:
            if key.startswith("baggage."):
                baggage_items[key[8:]] = span.attributes[key]

        return baggage_items

    def export(self, spans):
        try:
            for span in spans:
                baggage_items = self._extract_baggage_from_span(span)
                tenant_id = baggage_items.get("TenantId") or "miscellaneous"

                doc = self._build_document(span, baggage_items, tenant_id)
                self._queue.put_nowait((tenant_id, doc))
            return SpanExportResult.SUCCESS
        except Exception as ex:
            print(f"[MongoExporter] Export failed: {ex}")
            return SpanExportResult.FAILURE

    def _build_document(self, span, baggage_items, tenant_id: str):
        # Build ParentId in W3C trace context format
        if span.parent:
            parent_span_id = format(span.parent.span_id, "016x")
            parent_id = f"00-{format(span.context.trace_id, '032x')}-{parent_span_id}-01"
        else:
            parent_span_id = "0000000000000000"
            parent_id = ""
        
        return {
            "Timestamp": datetime.fromtimestamp(span.end_time / 1_000_000_000),
            "TraceId": format(span.context.trace_id, "032x"),
            "SpanId": format(span.context.span_id, "016x"),
            "ParentSpanId": parent_span_id,
            "ParentId": parent_id,
            "OperationName": span.name,
            "Kind": str(span.kind),
            "StartTime": datetime.fromtimestamp(span.start_time / 1_000_000_000),
            "EndTime": datetime.fromtimestamp(span.end_time / 1_000_000_000),
            "Duration": (span.end_time - span.start_time) / 1e6,
            "Attributes": {k: v for k, v in span.attributes.items() if not k.startswith("baggage.")},
            "Baggage": baggage_items,
            "Status": str(span.status.status_code),
            "StatusDescription": span.status.description,
            "ServiceName": self._service_name,
            "TenantId": tenant_id,
        }

    def _run(self):
        buffer_by_tenant = {}

        while not self._stop_event.is_set():
            try:
                tenant_id, doc = self._queue.get(timeout=self._flush_interval)
                buffer_by_tenant.setdefault(tenant_id, []).append(doc)

                while sum(len(docs) for docs in buffer_by_tenant.values()) < self._batch_size:
                    tenant_id, doc = self._queue.get_nowait()
                    buffer_by_tenant.setdefault(tenant_id, []).append(doc)
            except Empty:
                pass

            if buffer_by_tenant:
                self._flush_to_mongo(buffer_by_tenant)
                buffer_by_tenant.clear()

        self._flush_remaining()

    def _flush_to_mongo(self, batches):
        for tenant_id, docs in batches.items():
            try:
                self._db[tenant_id].insert_many(docs, ordered=False)
            except errors.PyMongoError as ex:
                print(f"[MongoExporter] Failed to insert docs for tenant '{tenant_id}': {ex}")

    def _flush_remaining(self):
        buffer_by_tenant = {}
        while not self._queue.empty():
            try:
                tenant_id, doc = self._queue.get_nowait()
                buffer_by_tenant.setdefault(tenant_id, []).append(doc)
            except Empty:
                break

        if buffer_by_tenant:
            self._flush_to_mongo(buffer_by_tenant)

    def force_flush(self, timeout_millis: int = 30000) -> bool:
        deadline = time.time() + (timeout_millis / 1000.0)
        while not self._queue.empty() and time.time() < deadline:
            time.sleep(0.1)
        self._flush_remaining()
        return True

    def shutdown(self):
        self._stop_event.set()
        self._worker_thread.join(timeout=self._flush_interval + 2)
        self._flush_remaining()
        self._client.close()



from dotenv import load_dotenv
import os
from typing import Dict, List

class EnvVaultConfig:
    @staticmethod
    def get_config(keys: List[str] = None) -> Dict[str, str]:
        load_dotenv()
        
        # Get all environment variables
        all_env = dict(os.environ)
        
        # # Print all environment variables
        # for key, value in all_env.items():
        #     print(f"{key}={value}")
        
        if keys:
            config = {key: os.environ.get(key) for key in keys}
            missing = [k for k, v in config.items() if not v]
            if missing:
                raise EnvironmentError(f"Missing environment variables: {', '.join(missing)}")
            return config
        
        return all_env



import asyncio
import json
import logging
from typing import Dict, List, Optional
from azure.servicebus.aio import ServiceBusClient, ServiceBusReceiver
from azure.servicebus import ServiceBusReceivedMessage
from opentelemetry import trace, baggage
from opentelemetry.trace import Status, StatusCode, SpanKind
from opentelemetry.trace.propagation.tracecontext import TraceContextTextMapPropagator
from blocks_genesis.auth.blocks_context import BlocksContext
from blocks_genesis.core.secret_loader import get_blocks_secret
from blocks_genesis.message.consumer import Consumer
from blocks_genesis.message.event_message import EventMessage
from blocks_genesis.message.message_configuration import MessageConfiguration

logger = logging.getLogger(__name__)

class AzureMessageWorker:
    def __init__(self, message_config: MessageConfiguration):
        self._logger = logger
        self._message_config = message_config
        self._consumer = Consumer()
        self._service_bus_client: Optional[ServiceBusClient] = None
        self._receivers: List[ServiceBusReceiver] = []
        self._active_message_renewals: Dict[str, asyncio.Event] = {}
        self._tracer = trace.get_tracer(__name__)

    def initialize(self):
        connection = self._message_config.connection or get_blocks_secret().MessageConnectionString
        if not connection:
            self._logger.error("Connection string missing")
            raise ValueError("Connection string missing")
        self._service_bus_client = ServiceBusClient.from_connection_string(connection)
        self._logger.info("✅ Service Bus Client initialized")

    async def stop(self):
        for message_id, event in self._active_message_renewals.items():
            event.set()
        self._active_message_renewals.clear()

        for receiver in self._receivers:
            try:
                await receiver.close()
            except Exception as ex:
                self._logger.error(f"Error closing receiver: {ex}")
        self._receivers.clear()

        if self._service_bus_client:
            await self._service_bus_client.close()

        self._logger.info("🛑 Worker stopped")

    async def run(self):
        if not self._service_bus_client:
            raise ValueError("Service Bus Client is not initialized")

        receiver_tasks = []

        # Process queues
        for queue_name in self._message_config.azure_service_bus_configuration.queues:
            receiver = self._service_bus_client.get_queue_receiver(
                queue_name=queue_name,
                prefetch_count=self._message_config.azure_service_bus_configuration.queue_prefetch_count
            )
            await receiver.__aenter__()  # Explicitly enter async context
            self._receivers.append(receiver)
            receiver_tasks.append(self.process_receiver(receiver))

        # Process topics
        for topic_name in self._message_config.azure_service_bus_configuration.topics:
            subscription_name = self._message_config.subscription_name.get(topic_name, "default-subscription")
            receiver = self._service_bus_client.get_subscription_receiver(
                topic_name=topic_name,
                subscription_name=subscription_name,
                prefetch_count=self._message_config.azure_service_bus_configuration.topic_prefetch_count
            )
            await receiver.__aenter__()
            self._receivers.append(receiver)
            receiver_tasks.append(self.process_receiver(receiver))

        self._logger.info("🚀 Receivers started")

        # This will block and keep running until cancelled
        await asyncio.gather(*receiver_tasks)

    async def process_receiver(self, receiver: ServiceBusReceiver):
        try:
            async for message in receiver:
                await self.message_handler(receiver, message)
        except asyncio.CancelledError:
            self._logger.info("Receiver cancelled")
        except Exception as ex:
            self._logger.error(f"Receiver error: {ex}")

    async def message_handler(self, receiver: ServiceBusReceiver, message: ServiceBusReceivedMessage):
        message_id = message.message_id
        self._logger.info(f"Received message: {message_id}")

        trace_id = message.application_properties.get(b"TraceId", b"").decode("utf-8")
        span_id = message.application_properties.get(b"SpanId", b"").decode("utf-8")
        tenant_id = message.application_properties.get(b"TenantId", b"").decode("utf-8")
        security_context_raw = message.application_properties.get(b"SecurityContext", b"").decode("utf-8")
        baggage_str = message.application_properties.get(b"Baggage", b"{}").decode("utf-8")

        if security_context_raw:
            BlocksContext.set_context(json.loads(security_context_raw))

        cancellation_event = asyncio.Event()
        self._active_message_renewals[message_id] = cancellation_event
        asyncio.create_task(self.start_auto_renewal_task(message, receiver, cancellation_event))

        try:
            context = TraceContextTextMapPropagator().extract({
                "traceparent": f"00-{trace_id}-{span_id}-01"
            })

            with self._tracer.start_as_current_span(
                "process.messaging.azure.service.bus",
                context=context,
                kind=SpanKind.CONSUMER
            ) as span:
                span.set_attribute("messaging.system", "azure.servicebus")
                span.set_attribute("message.id", message_id)
                span.set_attribute("SecurityContext", security_context_raw)
                span.set_attribute("message.body", str(message))
                baggages = json.loads(baggage_str)
                for key, value in baggages.items():
                    span.set_attribute(f"baggage.{key}", value)

                try:
                    start_time = asyncio.get_event_loop().time()
                    msg = EventMessage.parse_raw(message.body.decode("utf-8"))
                    await self._consumer.process_message(msg.type, msg.body)
                    processing_time = (asyncio.get_event_loop().time() - start_time) * 1000
                    self._logger.info(f"Processed message {message_id} in {processing_time:.1f}ms")

                    span.set_attribute("response", "Successfully Completed")
                    span.set_status(Status(StatusCode.OK, "Message processed successfully"))
                except Exception as ex:
                    self._logger.error(f"Error processing message {message_id}: {ex}")
                    span.set_status(Status(StatusCode.ERROR, str(ex)))
                    span.set_attribute("error", str(ex))
                    raise
                finally:
                    cancellation_event.set()
                    self._active_message_renewals.pop(message_id, None)
                    await receiver.complete_message(message)
                    self._logger.info(f"Message {message_id} completed.")
        except Exception as ex:
            self._logger.error(f"Unhandled error for message {message_id}: {ex}")
            cancellation_event.set()
            self._active_message_renewals.pop(message_id, None)
            await receiver.dead_letter_message(message, reason=str(ex))
        finally:
            BlocksContext.clear_context()

    async def start_auto_renewal_task(self, message: ServiceBusReceivedMessage, receiver: ServiceBusReceiver, cancellation_event: asyncio.Event):
        message_id = message.message_id
        start_time = asyncio.get_event_loop().time()
        renewal_count = 0
        renewal_interval = self._message_config.azure_service_bus_configuration.message_lock_renewal_interval_seconds
        max_processing_time = self._message_config.azure_service_bus_configuration.max_message_processing_time_minutes * 60

        try:
            while not cancellation_event.is_set():
                await asyncio.sleep(renewal_interval)
                processing_time = asyncio.get_event_loop().time() - start_time
                if processing_time > max_processing_time:
                    self._logger.warning(f"Message {message_id} exceeded max time ({max_processing_time}s); stopping lock renewal.")
                    break

                try:
                    await receiver.renew_message_lock(message)
                    renewal_count += 1
                    self._logger.info(f"Renewed lock for message {message_id} (#{renewal_count})")
                except Exception as ex:
                    self._logger.warning(f"Lock renewal failed for message {message_id}: {ex}")
                    break

            self._logger.info(f"Auto-renewal finished for {message_id} after {renewal_count} renewals")
        except asyncio.CancelledError:
            self._logger.info(f"Auto-renewal cancelled for {message_id} after {renewal_count} renewals")
        except Exception as ex:
            self._logger.error(f"Auto-renewal error for message {message_id}: {ex}")



from typing import Dict, Optional, List
import logging
from blocks_genesis.core.blocks_secret import BlocksSecret
from blocks_genesis.core.azure_key_vault import AzureKeyVault

logger = logging.getLogger(__name__)

# Module-level private variable to hold the immutable singleton secret
_loaded_secret: Optional[BlocksSecret] = None

class SecretLoader:
    def __init__(self, service_name: str = "blocks_service"):
        self.vault = AzureKeyVault()
        self.service_name = service_name

    async def load_secrets(self, fields: Optional[List[str]] = None) -> None:
        global _loaded_secret
        if _loaded_secret is not None:
            logger.debug("Secrets already loaded, skipping reload.")
            return

        fields = fields or list(BlocksSecret.model_fields.keys())
        try:
            logger.info("🔐 Loading secrets from Azure Key Vault...")
            raw_secrets: Dict[str, str] = await self.vault.get_secrets(fields)

            processed_secrets: Dict[str, object] = {
                key: (
                    value.lower() == "true"
                    if isinstance(value, str) and value.lower() in ["true", "false"]
                    else value
                )
                for key, value in raw_secrets.items()
            }

            secret = BlocksSecret(**processed_secrets)
            secret.ServiceName = self.service_name

            _loaded_secret = secret
            logger.info("✅ Secrets loaded successfully")
        except Exception:
            logger.exception("❌ Failed to load secrets")
            raise

    async def close(self) -> None:
        if hasattr(self.vault, "close") and callable(self.vault.close):
            await self.vault.close()

def get_blocks_secret() -> BlocksSecret:
    if _loaded_secret is None:
        raise Exception("Secrets not loaded")
    return _loaded_secret
